package com.arm.streamline;

import java.io.FileOutputStream;
import java.io.IOException;

public class StreamlineAnnotate {
	public static final Color				RED		= new Color(0xff, 0, 0);
	public static final Color				BLUE	= new Color(0, 0, 0xff);
	public static final Color				GREEN	= new Color(0, 0xff, 0);
	public static final Color				PURPLE	= new Color(0xff, 0, 0xff);
	public static final Color				YELLOW	= new Color(0xff, 0xff, 0);
	public static final Color				CYAN	= new Color(0, 0xff, 0xff);
	public static final Color				WHITE	= new Color(0xff);
	public static final Color				LTGRAY	= new Color(0xbb);
	public static final Color				DKGRAY	= new Color(0x55);
	public static final Color				BLACK	= new Color(0);
	private static final FileOutputStream	OUT;

	static {
		FileOutputStream out = null;
		try {
			out = new FileOutputStream("/dev/gator/annotate"); //$NON-NLS-1$
		} catch (Exception exception) {
			// Gator may not be installed. Ignore.
		}
		OUT = out;
	}

	/**
	 * Emit a textual annotation.
	 * 
	 * @param msg The message to emit.
	 */
	public static final void annotate(String msg) {
		annotate(null, msg);
	}

	/**
	 * Emit a colored textual annotation.
	 * 
	 * @param color The color to use.
	 * @param msg The message to emit.
	 */
	public static final void annotate(Color color, String msg) {
		if (OUT != null) {
			try {
				if (color != null) {
					OUT.write(0x1b);
					OUT.write(color.red);
					OUT.write(color.green);
					OUT.write(color.blue);
				}
				if (msg != null) {
					OUT.write(msg.getBytes());
				}
				OUT.write(0);
				OUT.flush();
				// Force the bytes to reach the file system, as flush() may not
				OUT.getFD().sync();
			} catch (IOException exception) {
				// Failed, but don't log
			}
		}
	}

	/**
	 * End an existing annotation. Rather than calling this method, you can instead emit a new
	 * annotation.
	 */
	public static final void end() {
		annotate(null, null);
	}

	/**
	 * Emit a visual annotation.
	 * 
	 * @param imageData The bytes of a supported image format: PNG, GIF, TIFF, JPEG, BMP.
	 */
	public static final void visualAnnotate(byte[] imageData) {
		visualAnnotate(null, null, imageData);
	}

	/**
	 * Emit a visual annotation with a colored text label.
	 * 
	 * @param msg The message to emit.
	 * @param imageData The bytes of a supported image format: PNG, GIF, TIFF, JPEG, BMP.
	 */
	public static final void visualAnnotate(String msg, byte[] imageData) {
		visualAnnotate(null, msg, imageData);
	}

	/**
	 * Emit a visual annotation with a colored text label.
	 * 
	 * @param color The color to use for the text.
	 * @param msg The message to emit.
	 * @param imageData The bytes of a supported image format: PNG, GIF, TIFF, JPEG, BMP.
	 */
	public static final void visualAnnotate(Color color, String msg, byte[] imageData) {
		if (OUT != null) {
			try {
				int msgLength;
				if (msg != null) {
					byte[] msgData = msg.getBytes();
					msgLength = msgData.length;
					if (color != null) {
						msgLength += 4;
					}
				} else {
					msgLength = 0;
				}
				byte[] header = new byte[8];
				header[0] = 0x1c;
				header[1] = 0x01;
				header[2] = (byte) (msgLength & 0xff);
				header[3] = (byte) (msgLength >> 8 & 0xff);
				int imgLength = imageData != null ? imageData.length : 0;
				header[4] = (byte) (imgLength & 0xff);
				header[5] = (byte) (imgLength >> 8 & 0xff);
				header[6] = (byte) (imgLength >> 16 & 0xff);
				header[7] = (byte) (imgLength >>> 24 & 0xff);
				OUT.write(header);
				if (msg != null) {
					if (color != null) {
						OUT.write(0x1b);
						OUT.write(color.red);
						OUT.write(color.green);
						OUT.write(color.blue);
					}
					OUT.write(msg.getBytes());
				}
				if (imgLength > 0) {
					OUT.write(imageData);
				}
				OUT.flush();
				// Force the bytes to reach the file system, as flush() may not
				OUT.getFD().sync();
			} catch (IOException exception) {
				// Failed, but don't log
			}
		}
	}
	
	/**
	 * Emit an annotation marker.
	 */
	public static final void marker() {
		marker(null, null);
	}
	
	/**
	 * Emit a textual annotation marker.
	 *
	 * @param msg The message to emit.
	 */
	public static final void marker(String msg) {
		marker(null, msg);
	}

	/**
	 * Emit a color annotation marker.
	 *
	 * @param color The color to use for the marker.
	 */
	public static final void marker(Color color) {
		marker(color, null);
	}
	
	/**
	 * Emit a color textual annotation marker.
	 *
	 * @param color The color to use for the marker.
	 * @param msg The message to emit.
	 */
	public static final void marker(Color color, String msg) {
		if (OUT != null) {
			try {
				OUT.write(0x1c);
				OUT.write(0x02);
				annotate(color, msg);
			} catch (IOException exception) {
				// Failed, but don't log
			}
		}
	}

	public static class Color {
		public int	red;
		public int	green;
		public int	blue;

		public Color(int redByte, int greenByte, int blueByte) {
			red = redByte;
			green = greenByte;
			blue = blueByte;
		}

		public Color(int gray) {
			this(gray, gray, gray);
		}
	}
}
